@@ -14,21 +14,46 @@
 #    under the License.
 
 import os
+import random
+
+import ddt
 
 import cinderlib
 from cinderlib.tests.functional import base_tests
 
 
-class BaseFunctTestCase(base_tests.unittest2.TestCase):
-    def test_list_supported_drivers(self):
+@ddt.ddt
+class BaseFunctTestCase(base_tests.unittest.TestCase):
+    @ddt.data([], [1], [2])
+    def test_list_supported_drivers(self, args):
+        is_v2 = args == [2]
+        expected_type = dict if is_v2 else str
         expected_keys = {'version', 'class_name', 'supported', 'ci_wiki_name',
                          'driver_options', 'class_fqn', 'desc'}
 
-        drivers = cinderlib.Backend.list_supported_drivers()
+        drivers = cinderlib.Backend.list_supported_drivers(*args)
         self.assertNotEqual(0, len(drivers))
         for name, driver_info in drivers.items():
             self.assertEqual(expected_keys, set(driver_info.keys()))
 
+            # Ensure that the RBDDriver has the rbd_keyring_conf option and
+            # it's not deprecated
+            if name == 'RBDDriver':
+                keyring_conf = [conf for conf in driver_info['driver_options']
+                                if conf['dest'] == 'rbd_keyring_conf']
+                self.assertEqual(1, len(keyring_conf))
+                expected_value = False if is_v2 else 'False'
+                self.assertEqual(expected_value,
+                                 keyring_conf[0]['deprecated_for_removal'])
+
+            for option in driver_info['driver_options']:
+                self.assertIsInstance(option['type'], expected_type)
+                if is_v2:
+                    self.assertTrue('type_class' in option['type'])
+                else:
+                    for v in option.values():
+                        self.assertIsInstance(v, str)
+
 
 @base_tests.test_all_backends
 class BackendFunctBasic(base_tests.BaseFunctTestCase):
@@ -48,31 +73,39 @@ class BackendFunctBasic(base_tests.BaseFunctTestCase):
         return sum(p.get('total_volumes', 0) for p in pools_info)
 
     def test_stats_with_creation(self):
-        # This test can fail if we are don't have exclusive usage of the
-        # storage pool used in the tests or if the specific driver does not
-        # return the right values in allocated_capacity_gb or
-        # provisioned_capacity_gb.
         initial_stats = self.backend.stats(refresh=True)
-        vol = self._create_vol(self.backend)
-        new_stats = self.backend.stats(refresh=True)
-
         initial_pools_info = self._pools_info(initial_stats)
-        new_pools_info = self._pools_info(new_stats)
-
         initial_volumes = self._volumes_in_pools(initial_pools_info)
-        new_volumes = self._volumes_in_pools(new_pools_info)
-
-        # If the backend is reporting the number of volumes, check them
-        if initial_volumes is not None:
-            self.assertEqual(initial_volumes + 1, new_volumes)
-
         initial_size = sum(p.get('allocated_capacity_gb',
                                  p.get('provisioned_capacity_gb', 0))
                            for p in initial_pools_info)
+
+        size = random.randint(1, 5)
+        vol = self._create_vol(self.backend, size=size)
+
+        # Check that without refresh we get the same data
+        duplicate_stats = self.backend.stats(refresh=False)
+        self.assertEqual(initial_stats, duplicate_stats)
+
+        new_stats = self.backend.stats(refresh=True)
+        new_pools_info = self._pools_info(new_stats)
+        new_volumes = self._volumes_in_pools(new_pools_info)
         new_size = sum(p.get('allocated_capacity_gb',
                              p.get('provisioned_capacity_gb', vol.size))
                        for p in new_pools_info)
-        self.assertEqual(initial_size + vol.size, new_size)
+
+        # We could be sharing the pool with other CI jobs or with parallel
+        # executions of this same one, so we cannot check that we have 1 more
+        # volume and 1 more GB used, so we just check that the values have
+        # changed.  This could still fail if another job just deletes 1 volume
+        # of the same size, that's why we randomize the size, to reduce the
+        # risk of the volumes having the same size.
+
+        # If the backend is reporting the number of volumes, check them
+        if initial_volumes is not None:
+            self.assertNotEqual(initial_volumes, new_volumes)
+
+        self.assertNotEqual(initial_size, new_size)
 
     def test_create_volume(self):
         vol = self._create_vol(self.backend)
